nil = b'\x00'


class Cache:
    active = True

    def __init__(self):
        self.cache = {}

    def clear(self, path=None):
        if path is None:
            return self.cache.clear()
        self.cache.pop(path, None)

    def get(self, key):
        if not self.active:
            return nil
        return self.cache.get(key, nil)

    def set(self, key, value):
        self.cache[key] = value


cache = Cache()



from .tools import die, env, log, need_env, run, exists, os
import requests
from .cache import cache, nil

E = need_env
base = 'https://api.hetzner.cloud/v1'

icons = {
    'networks': '🖧 ',
    'ssh_keys': '🔑',
    'servers': '🖥️',
    'load_balancers': '🌐',
    'images': '🐧',
    'volumes': '💾 ',
}


def headers():
    api_token = env('HCLOUD_TOKEN_WRITE')
    if not api_token:
        api_token = need_env('HCLOUD_TOKEN')
    return {'Authorization': f'Bearer {api_token}', 'Content-Type': 'application/json'}


def safe(r):
    try:
        return r.json()
    except Exception:
        return {'txt': r.text}


class hapi:
    def get(path):
        v = cache.get(path)
        if v != nil:
            return v
            # return log.debug('Cache hit', path=path) or v
        r = requests.get(f'{base}/{path}', headers=headers())
        r = cache.cache[path] = r.json()[path]
        return r

    def delete(path, id):
        log.info('HApi delete', path=path, id=id)
        r = requests.delete(f'{base}/{path}/{id}', headers=headers())
        if not r.status_code < 300:
            die('HApi delete failed', path=path, id=id, status_code=r.status_code)
        cache.clear(path)

    def post(path, data):
        i = icons.get(path, '')
        if i:
            log.info(f'{i} Create {path[:-1]}', **data)
        else:
            log.info('HApi post', path=path, **data)
        r = requests.post(f'{base}/{path}', headers=headers(), json=data)
        if not r.status_code < 300:
            die('HApi post failed', path=path, **safe(r))
        cache.clear(path)
        return r


def ips(name, no_die=False):
    S = by_name('servers', name)
    if not S:
        if no_die:
            return
        die('Server not found', name=name)
    ip = (S['public_net']['ipv4'] or {}).get('ip')
    priv = (S['private_net'][0] or {}).get('ip')
    return dict(pub=ip, priv=priv)


def by_name(typ, name, short=True, multi=False):
    N = E('NAME')
    if short and not name.startswith(N + '-'):
        name = E('NAME') + '-' + name
    r = [x for x in hapi.get(typ) if x['name'] == name]
    if r and len(r) > 1 and not multi:
        die(f'Multiple {typ} found', name=name)
    return r[0] if r else None



import structlog
import os
import sys
import json
import subprocess
import importlib.util

from rich.console import Console
from rich.prompt import Confirm
from .cache import nil
from .defaults import envdefaults
import time


console = Console()
now = lambda: int(time.time() * 1000)
T0 = now()


exists = os.path.exists


confirm = Confirm.ask

_secrets = {}

called = []


def shw(f, *a):
    origf = getattr(f, 'func', f)
    n = origf.__name__
    called.append(n)
    return log.info(f'⏺️ {n}') or f(*a)


class TemplRepl:
    add = {}

    def __getitem__(self, k, dflt=None):
        k, dflt = (k + '|').split('|')[:2]
        r = env(k, self.add.get(k, nil))
        if r == nil:
            if dflt is not None:
                return dflt
            die(f'Missing env var ${k}')
        return r


env_templ_repl = TemplRepl()


def render_env_into(tmpl, add=None):
    if add:
        env_templ_repl.add.update(add)
    tmpl = tmpl.replace('T(', '%(')
    tmpl = tmpl % env_templ_repl
    tmpl = tmpl.replace('True', 'true').replace('False', 'false')
    return tmpl + '\n'


def import_file(filename):
    d = os.path.dirname(os.path.abspath(filename))
    sys.path.append(d) if d not in sys.path else None
    return importlib.import_module(filename)


def pyval(k, fn, dflt=None):
    if not os.path.exists(f'{fn}.py'):
        die('Missing pyval file', k=k, fn=fn)
    mod = import_file(fn)
    v = getattr(mod, k, nil)
    if v == nil:
        if dflt is not None:
            return dflt
        die('Missing pyval var', k=k, fn=fn)
    return v


def env(key, dflt=None):
    v = os.environ.get(key, nil)
    if v == nil:
        v = getattr(envdefaults, key, dflt)
    if str(v).startswith('py:'):
        v = pyval(key, v[3:], dflt)
    elif str(v).startswith('pass:'):
        x = _secrets.get(v, nil)
        if x == nil:
            x = run(['pass', 'show', v[5:]], capture_output=True, text=True) or dflt
            _secrets[v] = x
        v = x
    return v


def dt(_, __, e):
    e['timestamp'] = f'{now() - T0:>4}'
    return e


structlog.configure(
    processors=[
        # structlog.processors.TimeStamper(fmt='iso'),
        dt,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.dev.ConsoleRenderer(),
    ],
    logger_factory=structlog.PrintLoggerFactory(file=sys.stderr),
)

log = structlog.get_logger()


def die(msg, **kw):
    log.fatal(msg, **kw)
    sys.exit(1)


def run(cmd, bg=False, **kw):
    if isinstance(cmd, str):
        cmd = cmd.split()
    pipe = kw.get('pipe', '')
    pipe = pipe if not len(pipe) > 20 else f'{pipe[:20]}...'
    log.debug('⚙️ Cmd', cmd=cmd, pipe=pipe)
    if bg:
        r = subprocess.Popen(cmd, start_new_session=True)
        # r.communicate()
        return r

    r = subprocess.run(cmd, **kw)
    if r.returncode != 0:
        die('Command failed', cmd=cmd, returncode=r.returncode)
    return r.stdout.strip() if r.stdout else ''


def need_env(k, dflt=None):
    v = env(k, dflt)
    if v is None:
        die(f'Missing env var ${k}')
    return v


def ssh(ip, port=None, cmd=None, input=None, send_env=None, capture_output=True, **kw):
    cmd = cmd if cmd is not None else 'bash -s'
    port = port if port is not None else int(env('SSH_PORT', 22))
    c = f'ssh -p {port} -o StrictHostKeyChecking=accept-new -i'.split()
    cmd = c + [need_env('FN_SSH_KEY'), f'root@{ip}', cmd]
    kw['capture_output'] = capture_output
    for key in send_env or []:
        cmd.insert(1, f'SendEnv={key}')
        cmd.insert(1, '-o')
    if cmd[-1] == 'args':  # for os.system style
        return [f'{e}' for e in cmd[1:-1]]
    kw['text'] = kw.get('text', True)
    return run(cmd, input=input, **kw)


def write_file(fn, s, log=0, mkdir=0, chmod=None, mode='w'):
    "API: Write a file. chmod e.g. 0o755 (as octal integer)"

    fn = os.path.abspath(fn)

    if log > 0:
        log.info('Writing file', fn=fn)

    if isinstance(s, (list, tuple)) and s and isinstance(s[0], str):
        s = '\n'.join(s)
    elif isinstance(s, (dict, tuple, list)):  # think of bytes, mode wb
        s = json.dumps(s, default=str)
    e = None
    for _ in 1, 2:
        try:
            with open(fn, mode) as fd:
                fd.write(s)
            if chmod:
                if not isinstance(chmod, (list, tuple)):
                    chmod = [int(chmod)]
                for s in chmod:
                    os.chmod(fn, s)
            return fn
        except IOError as ex:
            if mkdir:
                d = os.path.dirname(fn)
                os.makedirs(d)
                continue
            e = ex
        except Exception as ex:
            e = ex
        raise Exception('Could not write file: %s %s' % (fn, e))


def read_file(fn, dflt=None, mkfile=False, bytes=-1, strip_comments=False):
    """
    API function.
    read a file - return a default if it does not exist"""
    if not exists(fn):
        if dflt is not None:
            if mkfile:
                write_file(fn, dflt, mkdir=1)
            return dflt
        if mkfile:
            raise Exception(fn, 'Require dflt to make non existing file')
        raise Exception(fn, 'does not exist')
    with open(fn) as fd:
        # no idea why but __etc__hostname always contains a linesep at end
        # not present in source => rstrip(), hope this does not break templs
        res = fd.read(bytes)
        res = res if not res.endswith('\n') else res[:-1]
        if strip_comments:
            lines = res.splitlines()
            res = '\n'.join([l for l in lines if not l.startswith('#')])
        return res


env('DOMAIN')
env('DOMAIN')



import yaml

from .create import hk3s, tools, local
from .hapi import by_name, hapi, ips, need_env
from .tools import die, log, shw, ssh
from .ssh import port_forward, get_remote, run_remote

E = need_env
clearip = tools.clear_ip_from_known_hosts


def delete(name):
    """Deleting objects by name, via hapi"""
    S = by_name('servers', name)
    if not S:
        return log.info('Not found', name=name)
    return hapi.delete('servers', S['id'])


class recover:
    """When proxy died"""

    def kubeconfig():
        """rebuild a kubeconfig from the existing masters' k3s.yaml files"""
        y = []
        N = need_env('NAME')
        for i in range(1, 6):
            n = f'{N}-master{i}'
            ip = ips(n, no_die=True)
            if not ip:
                break
            kubecfg = get_remote(n, 'cat', '/etc/rancher/k3s/k3s.yaml')
            if not kubecfg:
                die('No k3s.yaml found on master', name=n)
            kubecfg = kubecfg.replace(
                'https://127.0.0.1:6443', f'https://{ip["priv"]}:6443'
            )
            kubecfg = kubecfg.replace('default', n)
            y.append(kubecfg)
        kc = [yaml.safe_load(i) for i in y]
        d = kc[0]
        for n in 'clusters', 'users', 'contexts':
            for k in kc[1:]:
                d[n].append(k[n][0])
        kubecfg = yaml.safe_dump(d)
        ip = ips('proxy')['pub']
        run_remote(ip, 'mkdir', '-p', '/root/.kube')
        ssh(ips('proxy')['pub'], cmd='tee /root/.kube/config', input=kubecfg)
        s = get_remote(ip, 'kubectl', 'get', 'nodes')
        if 'master1' not in s:
            die('Kubeconfig not working', output=s)
        print(s)
        log.info('Kubeconfig recovered')

    def hk3sconfig():
        """Rebuilt the config from environ and transfer to proxy"""
        shw(hk3s.recover_config)


class do:
    ssh = run_remote
    delete = delete
    download_kubectl = local.download_kubectl
    port_forward = port_forward



from .tools import (
    confirm,
    die,
    exists,
    log,
    need_env,
    os,
    read_file,
    render_env_into,
    run,
    ssh,
    write_file,
)
from .hapi import hapi, ips, cache
from .dns import dns_wildcard_add
import sys
import socket
import time
import json
import shutil
from .ssh import get_remote, kubecfg_fwd_port

from .assets.create_templ import T_INST_TOOLS_PROXY, T_CADDY, T_UNIT_FWD, T_SSHD

E = need_env

netname = lambda: f'ten-{E("HK_HOST_NETWORK")}'

here = os.path.dirname(__file__)


def linefeed():
    print('', file=sys.stderr)


class local:  # requirements
    def ensure_ssh_key_local():
        """Ensures local $FN_SSH_KEY pair exists, used when creating proxy"""
        fn = E('FN_SSH_KEY')
        if fn.endswith('.pub'):
            log.die('SSH key file should not end with .pub', fn=fn)
        fnp = fn + '.pub'
        os.makedirs(os.path.dirname(fn), exist_ok=True)
        if not exists(fn):
            log.warn('Creating SSH key', fn=fn)
            run(['ssh-keygen', '-q', '-t', 'ecdsa', '-N', '', '-f', fn])
        if not exists(fnp):
            run(['ssh-keygen', '-y', '-f', fn], stdout=fnp)
        os.chmod(fn, 0o600)
        log.debug('Have local SSH key', fn=fn, fnp=fnp)

    def download_kubectl():
        fn_dflt = E('HOME') + '/.kube/config'
        d = E('HOME') + '/.kube'
        fn = d + '/' + E('NAME')
        os.makedirs(d, exist_ok=True)
        c = get_remote('proxy', 'cat', '/root/.kube/config')
        if not c:
            die('No kubeconfig found on proxy')
        l = c.splitlines()
        c = []
        while l:
            c.append(l.pop(0))
            if c[-1].lstrip().startswith('server:'):
                line = c[-1].split(':')
                line[2] = '//127.0.0.1'
                line[3] = str(kubecfg_fwd_port)
                c[-1] = ':'.join(line)
        c = '\n'.join(c)
        o = read_file(fn, 'x')
        if os.path.islink(fn_dflt):
            os.unlink(fn_dflt)
        elif os.path.exists(fn_dflt):
            log.warn('Moving existing kubeconfig', fn_dflt=fn_dflt)
            shutil.move(fn_dflt, f'{fn_dflt}-{time.ctime()}')

        if o.strip() == c.strip():
            log.info('Kubeconfig already up to date')
        else:
            if 'https://' in o:
                fno = f'{fn}-{time.ctime()}'
                write_file(fno, o)
                log.info('backed up old kubeconfig', fn=fno)
            write_file(fn, c)
        os.symlink(fn, fn_dflt)
        log.info('have kubeconfig', fn=fn, symlinked=fn_dflt)


class proxy_:
    """Creates a Linux server acting as SSH proxy and LB - Only one with Pub IP"""

    def ensure_host_network():
        """Private network for all hosts"""
        nr = int(E('HK_HOST_NETWORK', '1'))
        r = hapi.get('networks')
        if netname() in [x['name'] for x in r]:
            return
        net = f'10.{nr}.0.0'
        sn = [{'ip_range': f'{net}/24', 'network_zone': 'eu-central', 'type': 'cloud'}]
        m = {
            'name': netname(),
            'ip_range': f'{net}/16',
            'expose_routes_to_vswitch': False,
            'subnets': sn,
            'labels': {'environment': 'dev', 'k3s': E('NAME')},
        }
        hapi.post('networks', data=m)

    def ensure_server():
        """Creates the server acting as SSH proxy and LB"""
        return tools.ensure_server(
            'proxy', E('HK_PROXY_TYPE'), E('HK_PROXY_IMG'), E('HK_LOCATION')
        )

    def ensure_default_route_via_proxy():
        """Ensures default route iof priv net is via proxy"""
        nets = hapi.get('networks')
        n = [x for x in nets if x['name'] == netname()]
        if not n:
            die('Network not found', name=netname())
        ip = ips('proxy')['priv']
        id = n[0]['id']
        routes = n[0]['routes']
        routes = [r for r in routes if r['destination'] == '0.0.0.0/0']
        d = lambda i: {'destination': '0.0.0.0/0', 'gateway': i}
        if routes:
            i = routes[0].get('gateway')
            if i == ip:
                return
            hapi.post(f'networks/{id}/actions/delete_route', data=d(i))
        hapi.post(f'networks/{id}/actions/add_route', data=d(ip))

    def ensure_is_ip_forwarder():
        """Persists and starts ip forwarding and NAT on proxy"""
        svc = 'ip_forwarder.service'
        ip = ips('proxy')['pub']
        have = ssh(ip, cmd='ls /etc/systemd/system')
        if svc in have:
            return
        u = T_UNIT_FWD(E('HK_HOST_NETWORK'))
        have = ssh(ip, cmd=f'tee /etc/systemd/system/{svc}', input=u)
        u = 'systemctl enable --now ip_forwarder'
        have = ssh(ip, input=u)

    def postinstall():
        """Installs kubectl, helm, kubectx, hetzner-k3s via binenv plus caddy/l4 proxy"""
        ip = ips('proxy')['pub']
        ssh(ip, input=T_SSHD)
        t = render_env_into(T_INST_TOOLS_PROXY)
        ssh(ip, input=t, capture_output=False)
        url = E('URL_CADDY')
        ssh(ip, cmd=f'curl -s -L -o /root/caddy "{url}"', bg=True, capture_output=False)

    def configure_caddy():
        """Configures Caddy as a Layer 4 proxy into the cluster, incl. proxy protocol v2"""
        s = {}
        for p in [80, 443]:
            # 🧧 its a service and we use a secure master one. works but not optimal:
            dials = [f'10.{E("HK_HOST_NETWORK")}.0.5:{p + 30000}']
            h = {
                'handler': 'proxy',
                'proxy_protocol': 'v2',
                'upstreams': [{'dial': dials}],
            }
            s[f'port{p}'] = {
                'listen': [f':{p}', f'[::]:{p}'],
                'routes': [{'handle': [h]}],
            }
        c = {
            'logging': {
                'sink': {'writer': {'output': 'stdout'}},
                'logs': {'default': {'level': 'DEBUG'}},
            },
            'apps': {'layer4': {'servers': s}},
        }
        ssh(ips('proxy')['pub'], input=T_CADDY(json.dumps(c)), capture_output=False)


class tools:
    def ensure_ssh_key_known_to_hetzner():
        have = hapi.get('ssh_keys') or []
        fps = set([k['fingerprint'] for k in have])
        fp = E('FN_SSH_KEY') + '.pub'
        r = run(
            ['ssh-keygen', '-l', '-E', 'md5', '-f', fp],
            capture_output=True,
            text=True,
        )
        r = r.split(':', 1)[1].split()[0]
        if r in fps:
            return [k['name'] for k in have if k['fingerprint'] == r][0]
        n = E('NAME')
        if n in [k['name'] for k in have]:
            msg = f'proceed to delete and re-add the SSH key {n}'
            if not confirm(msg, default=False):
                die('unconfirmed')
            f = [k['id'] for k in have if k['name'] == n]
            r = hapi.delete('ssh_keys', f[0])
        r = hapi.post('ssh_keys', data={'name': n, 'public_key': open(fp).read()})
        return n

    def ensure_server(shortname, typ, img, loc):
        key_name = tools.ensure_ssh_key_known_to_hetzner()
        name = E('NAME') + '-' + shortname
        have = hapi.get('servers') or []
        if name not in [x['name'] for x in have]:
            nid = [x['id'] for x in hapi.get('networks') if x['name'] == netname()]
            if not nid:
                die('No network found', name=netname())
            data = dict(
                name=name,
                server_type=typ,
                image=img,
                location=loc,
                ssh_keys=[key_name],
                networks=nid,
                public_net=dict(enable_ipv4=True, enable_ipv6=True),
                labels=dict(environment='dev', k3s=E('NAME')),
            )
            while True:
                server = hapi.post('servers', data=data)
                srv_id = server.json()['server']['id']
                time.sleep(2)
                ip = ips(shortname)['pub']

                log.info('Server ordered - waiting for ssh 🕐', name=name, ip=ip)
                tools.clear_ip_from_known_hosts(ip)
                if not tools.wait_ssh_port(ip, 22):
                    hapi.delete('servers', srv_id)
                    continue
                linefeed()
                log.info('Server is up - waiting for ip priv', name=name, ip=ip)
                cache.active = False
                while True:
                    p = ips(shortname)['priv']
                    if p:
                        log.info(f'have priv ip {p}')
                        break
                    time.sleep(2)
                cache.active = True
                r = ssh(ip, port=22, cmd='curl -s ipinfo.io/ip')
                linefeed()
                if r == ip:
                    break
                msg = f'Tough luck - our ip {ip} seems to be not acceptable by some internet services. Trying again.'
                log.warn(msg)
                hapi.delete('servers', srv_id)

        log.info('Server is up', name=name, **ips(shortname))

    def wait_ssh_port(host, port, max_retries=150, dt=2):
        buffer = ''
        for _ in range(max_retries):
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.settimeout(dt)
                    s.connect((host, port))
                    while True:
                        data = s.recv(1024).decode('utf-8')
                        if not data:
                            break
                        buffer += data
                        if 'SSH' in buffer:
                            return True
            except (socket.error, socket.timeout) as _:
                print('.', end='', flush=True, file=sys.stderr)
                time.sleep(dt)
        msg = f'Failed to find "SSH" in the reply from {host}.'
        log.warning(msg, timeout=dt * max_retries)
        return False

    def clear_ip_from_known_hosts(ip):
        ip = ip + ' '
        fn = os.path.expanduser('~/.ssh/known_hosts')
        h = read_file(fn, dflt='')
        if ip not in h:
            return
        hn = [l for l in h.splitlines() if not l.startswith(ip)]
        if hn != h:
            write_file(fn, '\n'.join(hn), mkdir=1)


class hk3s:
    def render_config():
        """required standalone for recovering the proxy"""
        ip = ips('proxy', no_die=True)
        key_proxy = '# </root/.ssh/id_ed25519.pub of proxy server - not created yet>'
        if ip:
            ip = ip['pub']
            key_proxy = ssh(ip, cmd='cat /root/.ssh/id_ed25519.pub')
        dtmpl = f'{here}/assets'
        tmpl = E('HK_CFG_TMPL', '') or f'{dtmpl}/hk3s.yml'
        pmpl = E('HK_CFG_POST', '') or f'{dtmpl}/hk3s_post_create.yml'
        cfg = render_env_into(read_file(tmpl))
        d = dict(key_proxy=key_proxy, key_local=read_file(E('FN_SSH_KEY') + '.pub'))
        return cfg + render_env_into(read_file(pmpl), add=d)

    def recover_config():
        ip = ips('proxy')['pub']
        cfg = hk3s.render_config()
        ssh(ip, cmd='tee /root/config.yml', input=cfg)
        return ip

    def install():
        """Renders hetzner-k3s config, copies to proxy and runs the tool"""
        hk3s.recover_config()
        ip = ips('proxy')['pub']
        v = os.environ.get('HCLOUD_TOKEN')
        os.environ['HCLOUD_TOKEN'] = E('HCLOUD_TOKEN_WRITE')
        cmd = 'kubectl get nodes 2>/dev/null && echo "Skipping install - already running" || '
        cmd += 'hetzner-k3s create --config /root/config.yml'
        ssh(ip, cmd=cmd, capture_output=False, send_env=['HCLOUD_TOKEN'])
        os.environ['HCLOUD_TOKEN'] = v


class dns:
    def add_subdomain():
        """Points a whole subdomain to the proxy server"""
        return dns_wildcard_add(ips('proxy')['pub'])


class create:
    """Create New Infrastructure"""

    proxy = [
        local.ensure_ssh_key_local,
        proxy_.ensure_host_network,
        proxy_.ensure_server,
        proxy_.ensure_default_route_via_proxy,
        proxy_.ensure_is_ip_forwarder,
        proxy_.postinstall,
    ]
    k3s = hk3s.install
    proxylb = proxy_.configure_caddy
    dns = dns.add_subdomain


if __name__ == '__main__':
    main()



T_UNIT_FWD = (
    lambda net: f"""
[Unit]
Description=Bastion IP Forwarder

[Service]
Type=oneshot
ExecStart=/bin/bash -c 'echo 1 > /proc/sys/net/ipv4/ip_forward'
ExecStart=/bin/bash -c 'iptables -t nat -A POSTROUTING -s "10.{net}.0.0/16" -o eth0 -j MASQUERADE'

[Install]
WantedBy=multi-user.target
"""
)


T_SSHD = """
grep -q "HCLOUD_TOKEN" /etc/sshd/sshd_config || {
    echo 'AcceptEnv HCLOUD_TOKEN' >> /etc/ssh/sshd_config
    echo 'PasswordAuthentication no' >> /etc/ssh/sshd_config
    echo 'ClientAliveInterval 60' >> /etc/ssh/sshd_config
    echo 'ClientAliveCountMax 4' >> /etc/ssh/sshd_config
    systemctl daemon-reload
    systemctl reload sshd || systemctl restart ssh
}
type binenv 2>/dev/null || sed -i '1iexport PATH="$HOME/.binenv:$PATH"' ~/.bashrc
test -e "/root/.ssh/id_ed25519" || ssh-keygen -q -t ecdsa -N '' -f "/root/.ssh/id_ed25519" >/dev/null
touch /etc/postinstalled
"""


T_INST_TOOLS_PROXY = """
    function have_ { type "$1" 2>/dev/null ; }
    function install_binenv {
        echo "Installing $binenv"
        wget -q "https://raw.githubusercontent.com/axgkl/binaries/refs/heads/master/binenv-install.sh" -O - | bash
        have_ "binenv" || { echo "binenv install failed" && exit 1; }
    }
    export URL_BINENV_DISTRIS="%(URL_BINENV_DISTRIS|)s"
    export BINENV_TOOLS="%(BINENV_TOOLS_PROXY|)s %(BINENV_ADD_TOOLS_PROXY|)s"
    have_ "binenv" || install_binenv
    eval "binenv install $BINENV_TOOLS"
"""


T_CADDY = (
    lambda cfg: f"""
useradd -m -d /home/caddy caddy || true
mkdir -p /opt/caddy
test -f /root/caddy && mv /root/caddy /opt/caddy/caddy
chmod +x /opt/caddy/caddy
echo -e '{cfg}' > /opt/caddy/config.json
chown -R caddy:caddy /opt/caddy
echo -e '
[Unit]
Description=Caddy Web Server
Documentation=https://caddyserver.com/docs/
After=network.target

[Service]
User=caddy
ExecStartPre=/usr/bin/mkdir -p /home/caddy/.config
ExecStart=/opt/caddy/caddy run --config /opt/caddy/config.json
TimeoutStopSec=5s
LimitNOFILE=1048576
LimitNPROC=512
PrivateTmp=true
ProtectSystem=full
AmbientCapabilities=CAP_NET_BIND_SERVICE

[Install]
WantedBy=multi-user.target
' > /etc/systemd/system/caddy.service
systemctl daemon-reload
for a in enable restart; do systemctl $a caddy; done
systemctl status caddy
"""
)



from ipaddress import ip_address
from .hapi import ips
from .tools import log, ssh, need_env as E
from functools import partial
import sh


def ips_of_host(name):
    """name either ip or hostname"""
    try:
        if ip_address(name).is_private:
            priv = name
            pub = priv = ips('proxy')['pub']
        else:
            priv = pub = name
    except ValueError:
        priv = pub = ips(name)['pub']
        if not priv:
            pub = priv = ips('proxy')['pub']
            priv = ips(name)['priv']
    return pub, priv


def ssh_add_no_hostkey_check(args):
    args.insert(0, 'UserKnownHostsFile=/dev/null')
    args.insert(0, '-o')
    shc = 'StrictHostKeyChecking'
    for n, a in zip(range(len(args)), args, strict=True):
        if a.startswith(shc):
            args[n] = f'{shc}=no'
            return


def port_forward():
    fwd = f'{kubecfg_fwd_port}:127.0.0.1:6443'
    run_remote('master1', 'htop', '-d', '50', _term=True, _fwd=fwd)


def run_remote(name, *cmd, _fwd=None, _term=False, _fg=True):
    """ssh to servers, e.g. ssh proxy [cmd]. autovia via proxy built in."""
    log.debug('Run remote', name=name, cmd=cmd)
    ip_pub, ip = ips_of_host(name)
    args = ssh(ip_pub, cmd='args')
    if ip != ip_pub:
        ssh_add_no_hostkey_check(args)
        args.insert(-1, '-J')
        args.append(f'root@{ip}')
        # clearip(ip)
    if _term:
        args.insert(0, '-t')
    if _fwd:
        args.insert(0, _fwd)
        args.insert(0, '-L')
    args.extend(list(cmd))
    return sh.ssh(args, _fg=_fg)  # this can be redirected w/o probs


kubecfg_fwd_port = int(E('HK_HOST_NETWORK')) + 6443

get_remote = partial(run_remote, _fg=False)



from .tools import die, need_env, log
import requests


class providers:
    class digitalocean:
        @classmethod
        def api(cls, pth, meth=requests.get, data=None):
            basedomain = need_env('DOMAIN').split('.', 1)[1]  # axlc.net
            token = need_env('DNS_API_TOKEN')
            _ = 'application/json'
            headers = {'Content-Type': _, 'Authorization': f'Bearer {token}'}
            url = f'https://api.digitalocean.com/v2/domains/{basedomain}/{pth}'
            if data:
                r = meth(url, headers=headers, json=data)
            else:
                r = meth(url, headers=headers)
            if not r.status_code < 300:
                die('digitalocean API error', status=r.status_code, url=url)
            if not meth == requests.delete:
                return r.json()

        @classmethod
        def dns_wildcard_add(cls, ip, dom, subdom):
            l = [e for e in cls.list() if e['name'] == f'*.{subdom}']
            if l:
                if l[0]['data'] == ip:
                    return log.info('DNS wildcard already set', ip=ip, subdom=subdom)
                cls.api(pth=f'records/{l[0]["id"]}', meth=requests.delete)
            ttl = int(need_env('DNS_TTL', 60))
            d = dict(type='A', name=f'*.{subdom}', data=ip, ttl=ttl)
            r = cls.api(pth='records', meth=requests.post, data=d)
            log.info('DNS wildcard added', **r['domain_record'])

        @classmethod
        def list(cls):
            r = cls.api('records').get('domain_records', [])
            log.debug(
                'DigitalOcean records',
                records=[e['name'] for e in r if not e['name'] == '@'],
            )
            return r


def dns_wildcard_add(ip):
    prov = need_env('DNS_PROVIDER')
    wildcard_add = getattr(providers, prov)
    if not wildcard_add:
        return log.error('Unknown DNS provider', provider=prov)
    wildcard_add = wildcard_add.dns_wildcard_add
    dom = need_env('DOMAIN')
    l = dom.split('.')
    subdom, dom = l[0], '.'.join(l[1:])
    return wildcard_add(ip, dom, subdom)



binenv_custom_base = 'https://github.com/axgkl/binaries/raw/master'
dflt_img = 'ubuntu-24.04'
dflt_type = 'cx22'


class envdefaults:
    BINENV_TOOLS_PROXY = 'helm 3.16.3 kubectl 1.31.3 hetzner-k3s 2.2.3 btop 1.4.0'
    DNS_API_TOKEN = 'your-token-to-add-a-wildcard-dns-entry-at-provider'
    DNS_PROVIDER = 'digitalocean'  # see dns.py for others
    DNS_TTL = 60
    DOMAIN = 'k8s.mycompany.net'
    EMAIL = 'me@mycompany.com'
    FN_SSH_KEY = '$HOME/.ssh/hetzner-cluster'  # created if not exists, also on hcloud
    GITOPS_BRANCH = 'main'
    GITOPS_HOST = 'gitlab.mycompany.com'
    GITOPS_OWNER = 'company'
    GITOPS_PATH = 'clusters/staging'
    GITOPS_REPO = 'k8s'
    GITOPS_TOKEN = ''
    HCLOUD_TOKEN = ''
    HCLOUD_TOKEN_WRITE = ''
    HK_AUTOSCL_COUNT = 9
    HK_AUTOSCL_IMG = dflt_img
    HK_AUTOSCL_TYPE = 'cpx31'  # expensive
    HK_CIDR_CLUSTER = '10.50.0.0/16'
    HK_CIDR_SERVICE = '10.60.0.0/16'
    HK_CNI = 'cilium'
    HK_DNS_CLUSTER = '10.60.0.10'
    HK_HOST_NETWORK = 1
    HK_LOCATION = 'hel1'
    HK_MASTERS_ARE_WORKERS = True
    HK_MASTERS_COUNT = 3
    HK_MASTERS_IMG = dflt_img
    HK_MASTERS_TYPE = dflt_type
    HK_PROXY_IMG = 'ubuntu-24.04'
    HK_PROXY_TYPE = 'cx22'
    HK_REGISTRY_MIRROR = True
    HK_CFG_TMPL = ''  # empty: default template
    HK_CFG_POST = ''
    HK_VER = '2.2.3'
    HK_VER_K3S = 'v1.30.2+k3s2'
    HK_WORKERS_COUNT = 0
    HK_WORKERS_IMG = dflt_img
    HK_WORKERS_TYPE = dflt_type
    NAME = 'k3s'
    URL_BINENV_DISTRIS = f'{binenv_custom_base}/distributions.patch.yaml'
    URL_CADDY = f'{binenv_custom_base}/caddy-amd64'



import os, sys
import json
from .tools import T0, die, log, now, shw, called
from inspect import signature as sig
from functools import partial
from .create import create, hk3s
from .do import do, recover
from rich.console import Console
from rich.tree import Tree

console = Console()


B = lambda a: f'\x1b[1m{a}\x1b[0m'
L = lambda a: f'\x1b[2m{a}\x1b[0m'
D = lambda a: L((getattr(a, '__doc__', '') or '').strip().split('\n', 1)[0])


def tc(cls):
    return f'{B(cls.__name__)} {D(cls)}'


def tv(fnc):
    return f'{fnc.__name__} {D(fnc)}'


def cls_help(cls, inittree=None):
    tree = inittree or Tree(tc(cls))
    for k in [e for e in dir(cls) if not e.startswith('_')]:
        v = getattr(cls, k)
        if isinstance(v, type):
            tr = tree.add(f'{B(k)} {D(v)}')
            cls_help(v, tr)
            continue
        if not isinstance(v, list):
            tree.add(f'{B(k)} {tv(v)}')
            continue
        tr = tree.add(B(k))
        for m in v:
            tr.add(tv(m))
    if not inittree:
        console.print(tree)
        sys.exit(0)


class pyhk3:
    create = create
    do = do
    recover = recover
    hk3s = hk3s


def create_partial(funcs, args, argv, x):
    """the last func in a argv list may have arguments which we map here to a partial of that func"""
    args.pop(0)
    l = funcs
    if isinstance(l[-1], list):
        l = l[-1]
    kw, varg = {}, ()
    for _, p in sig(l[-1]).parameters.items():
        if int(p.kind) == 1:  # positional_or_keyword
            varg += (args.pop(0),)
        elif int(p.kind) == 2:  # vararg
            varg += tuple(args)
            args.clear()
    if not args:
        l[-1] = partial(l[-1], *varg, **kw)
        return zip(argv, funcs, strict=True)
    die('Unsupported', cmd=x)


def funcs(cls, args):
    funcs = []
    argv = []
    for x in args:
        f = getattr(cls, x, None)
        if not f:
            if not funcs:
                die('Unsupported', cmd=x)
            return create_partial(funcs, args, argv, x)
        funcs.append(f)
        argv.append(x)
    return zip(argv, funcs, strict=True)


def run_cls(cls, args):
    if not args or args[0] in {'help', '--help', '-h'}:
        cls_help(cls)

    for n, f in funcs(cls, args):
        log.info(f'🟧 {n}')
        if not isinstance(f, list):
            f = [f]
        for g in f:
            lastr = shw(g)
    acts = ' '.join(called)
    log.info(f'🟩 Done: {acts} [{round((now() - T0) / 1000.0, 2)}s]')
    return lastr


def main():
    if not len(sys.argv) > 1 or sys.argv[1] in {'help', '--help', '-h'}:
        cls_help(pyhk3)

    a = sys.argv[1]
    c = getattr(pyhk3, a, None)
    if c is None:
        die('Unsupported', cmd=a)
    r = run_cls(c, sys.argv[2:])
    print_result(r)


def print_result(r):
    if not r:
        return
    if isinstance(r, (dict, list, tuple)):
        try:
            r = json.dumps(r, indent=2, default=str)
        except Exception:
            pass
    print(r)


if __name__ == '__main__':
    main()
